"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.MerkleDistributorWrapper = void 0;
const solana_contrib_1 = require("@saberhq/solana-contrib");
const token_utils_1 = require("@saberhq/token-utils");
const spl_token_1 = require("@solana/spl-token");
const web3_js_1 = require("@solana/web3.js");
const pda_1 = require("./pda");
const utils_1 = require("./utils");
class MerkleDistributorWrapper {
    constructor(sdk, key, distributorATA, data) {
        this.sdk = sdk;
        this.program = sdk.program;
        this.key = key;
        this.distributorATA = distributorATA;
        this.data = data;
    }
    static async load(sdk, key) {
        const data = await sdk.program.account.merkleDistributor.fetch(key);
        return new MerkleDistributorWrapper(sdk, key, await (0, token_utils_1.getATAAddress)({ mint: data.mint, owner: key }), data);
    }
    static async createMerkleAirdrop(args) {
        const { airdropList, airdropName, tokenMint, sdk } = args;
        const balanceMap = {};
        airdropList.forEach(({ authority, amount }) => {
            const prevBalance = balanceMap[authority];
            if (prevBalance) {
                balanceMap[authority] = prevBalance.add(new token_utils_1.u64(amount));
            }
            else {
                balanceMap[authority] = new token_utils_1.u64(amount);
            }
        });
        const { claims, merkleRoot, tokenTotal } = (0, utils_1.parseBalanceMap)(Object.entries(balanceMap).map(([authority, amount]) => ({
            address: authority,
            earnings: amount.toString(),
        })));
        const pendingDistributor = await this.createDistributor({
            sdk,
            root: merkleRoot,
            maxTotalClaim: new token_utils_1.u64(tokenTotal),
            maxNumNodes: new token_utils_1.u64(Object.keys(claims).length),
            tokenMint: tokenMint ? tokenMint : token_utils_1.NATIVE_MINT,
            name: airdropName,
            withdrawAuthority: sdk.provider.wallet.publicKey,
        });
        return pendingDistributor;
    }
    static async createDistributor(args) {
        const { root, tokenMint, name, withdrawAuthority } = args;
        // Validate the name must be less than 32 bytes
        if (Buffer.from(name).length > 32) {
            throw new Error("Name must be less than 32 bytes");
        }
        const { sdk } = args;
        const { provider } = sdk;
        const [distributor, bump] = await (0, pda_1.findDistributorKey)(provider.wallet.publicKey, name);
        const ixs = [];
        ixs.push(sdk.program.instruction.newDistributor(bump, name, (0, utils_1.toBytes32Array)(root), args.maxTotalClaim, args.maxNumNodes, withdrawAuthority, {
            accounts: {
                base: provider.wallet.publicKey,
                distributor,
                mint: tokenMint,
                payer: provider.wallet.publicKey,
                systemProgram: web3_js_1.SystemProgram.programId,
            },
        }));
        const { address, instruction } = await (0, token_utils_1.getOrCreateATA)({
            provider,
            mint: tokenMint,
            owner: distributor,
        });
        if (instruction) {
            ixs.push(instruction);
        }
        return {
            creator: provider.wallet.publicKey,
            name,
            bump,
            distributor,
            distributorATA: address,
            root: root.toString("hex"),
            tx: new solana_contrib_1.TransactionEnvelope(provider, ixs),
        };
    }
    async claimIX(args, payer) {
        const { amount, claimant, index, proof } = args;
        const [claimStatus, bump] = await (0, pda_1.findClaimStatusKey)(index, this.key);
        return this.program.instruction.claim(bump, index, amount, proof.map((p) => (0, utils_1.toBytes32Array)(p)), {
            accounts: {
                distributor: this.key,
                claimStatus,
                from: this.distributorATA,
                to: await (0, token_utils_1.getATAAddress)({ mint: this.data.mint, owner: claimant }),
                claimant,
                payer,
                systemProgram: web3_js_1.SystemProgram.programId,
                tokenProgram: token_utils_1.TOKEN_PROGRAM_ID,
            },
        });
    }
    async claim(args) {
        const { provider } = this.sdk;
        const tx = new solana_contrib_1.TransactionEnvelope(provider, [
            await this.claimIX(args, provider.wallet.publicKey),
        ]);
        const { instruction } = await (0, token_utils_1.getOrCreateATA)({
            provider,
            mint: this.data.mint,
            owner: args.claimant,
        });
        if (instruction) {
            tx.instructions.unshift(instruction);
        }
        // check data.mint is wrapped sol or not
        // if wrapped sol, add unwrap instruction
        if (this.data.mint.equals(token_utils_1.NATIVE_MINT)) {
            const tokenAccount = await spl_token_1.Token.getAssociatedTokenAddress(token_utils_1.ASSOCIATED_TOKEN_PROGRAM_ID, token_utils_1.TOKEN_PROGRAM_ID, token_utils_1.NATIVE_MINT, provider.wallet.publicKey);
            const unwrapIx = spl_token_1.Token.createCloseAccountInstruction(token_utils_1.TOKEN_PROGRAM_ID, tokenAccount, provider.wallet.publicKey, provider.wallet.publicKey, []);
            tx.instructions.push(unwrapIx);
        }
        return tx;
    }
    async getClaimStatus(index) {
        const [key] = await (0, pda_1.findClaimStatusKey)(index, this.key);
        return this.program.account.claimStatus.fetch(key);
    }
    async reload() {
        this.data = await this.program.account.merkleDistributor.fetch(this.key);
    }
}
exports.MerkleDistributorWrapper = MerkleDistributorWrapper;
//# sourceMappingURL=wrapper.js.map