import { newProgram } from "@saberhq/anchor-contrib";
import {
  type AugmentedProvider,
  type Provider,
  SignerWallet,
  SingleConnectionBroadcaster,
  SolanaProvider,
} from "@saberhq/solana-contrib";
import { SolanaAugmentedProvider } from "@saberhq/solana-contrib";
import {
  type Keypair,
  type PublicKey,
  type Signer,
  Connection,
} from "@solana/web3.js";

import { PROGRAM_ID } from "./constants";
import { MerkleDistributorJSON } from "./idls/merkle_distributor";
import type {
  CreateAirdropArgs,
  CreateDistributorArgs,
  MerkleDistributorProgram,
  PendingDistributor,
} from "./types";
import { MerkleDistributorWrapper } from "./wrapper";

export class MerkleDistributorSDK {
  constructor(
    readonly provider: AugmentedProvider,
    readonly program: MerkleDistributorProgram
  ) {}

  withSigner(signer: Signer): MerkleDistributorSDK {
    return MerkleDistributorSDK.load({
      provider: this.provider.withSigner(signer),
    });
  }

  static new(rpcURL: string, keypair: Keypair): MerkleDistributorSDK {
    const connection = new Connection(rpcURL);

    const provider = new SolanaProvider(
      connection,
      new SingleConnectionBroadcaster(connection),
      new SignerWallet(keypair)
    );
    return MerkleDistributorSDK.load({ provider });
  }

  /**
   * Loads the SDK.
   * @returns {MerkleDistributorSDK}
   */
  static load({
    provider,
  }: {
    // Provider
    provider: Provider;
  }): MerkleDistributorSDK {
    const aug = new SolanaAugmentedProvider(provider);
    return new MerkleDistributorSDK(
      aug,
      newProgram<MerkleDistributorProgram>(
        MerkleDistributorJSON,
        PROGRAM_ID,
        aug
      )
    );
  }

  /**
   * Load an existing merkle distributor.
   * @returns {MerkleDistributorWrapper}
   */
  async loadDistributor(key: PublicKey): Promise<MerkleDistributorWrapper> {
    return await MerkleDistributorWrapper.load(this, key);
  }

  /**
   * Create a merkle distributor.
   * @returns {PendingDistributor}
   */
  async createDistributor(
    args: Omit<CreateDistributorArgs, "sdk">
  ): Promise<PendingDistributor> {
    return await MerkleDistributorWrapper.createDistributor({
      sdk: this,
      ...args,
    });
  }

  async createMerkleAirdrop(
    args: Omit<CreateAirdropArgs, "sdk">
  ): Promise<PendingDistributor> {
    return await MerkleDistributorWrapper.createMerkleAirdrop({
      sdk: this,
      ...args,
    });
  }
}
